;-*- mode:LISP; Package:AP5; Base:10; Syntax: Common-lisp -*-

; code for dumping dbobjects and memo cells

(in-package "AP5")
#| 
It would sure be nice if "dumping" were generic
so we could define how to dump instances of a type where
we define the type.  But it isn't, so we have to define
all the types we want to dump specially before the dumping
code.
|#

(defstruct (DBObject ; :named  (:type vector)
	     (:conc-name nil)
	     (:print-function dbobject-print))
  classes Properties)

(proclaim '(special bootstrapping))
; since I don't want to let anyone unadvise this (, Dennis!), I won't do it via advice.
(defvar *no-query-dump-dbobjects* nil
  "When T, new-dump-array will not complain if it is dumping DBOBJECTS")
(defvar *no-query-dump-memos* nil
  "When T, new-dump-array will not complain if it is dumping memos")
#+symbolics
(progn
(unless (fboundp 'old-dump-array)
  (setf (symbol-function 'old-dump-array) (symbol-function 'si:dump-array)))

(defun new-dump-array (array stream)
  (cond ((dbobject-p array)
	 (unless (relationp array)
	   (if *no-query-dump-dbobjects* (push array *no-query-dump-dbobjects*)
	       (cerror "Proceed to dump ~a."
		       "SI:DUMP-ARRAY called on a dbobject (~a).~%~
                 This will work on your symbolics machine, but your code is not portable."
		       array)))
	 (si:DUMP-FORM-TO-EVAL
	   (cond ((relationp array)
		   `(relationp ',(relationnamep array)))
		  (t `(read-from-string
			,(let ((*print-escape* t))
			   (PrintDBObject array)))))
	   stream))
	((memodata-p array)
	 (when (done array)
	   (if *no-query-dump-memos* (push array *no-query-dump-memos*)
	     (cerror "Proceed to dump the memo cell."
		     "DUMP-ARRAY called on a loaded memo cell.~%~
                     This will work on your symbolics machine, but your code is not portable.")))
	    ; it looks like we do indeed expect to compile memos to files
	    ; however they evidently are never dumped with done=T
	  (si:DUMP-FORM-TO-EVAL '(make-memodata) stream))
	((loadmemodata-p array)
	 (si:DUMP-FORM-TO-EVAL (lm-form array) stream))
	(t (old-dump-array array stream))))

(setf (symbol-function 'si::dump-array) (symbol-function 'new-dump-array))
)

; ----------------

#+ ignore ;ti
;; ti does not need to dump dbobjects or memo cells
(progn
(unless (fboundp 'old-dump-array)
  (setf (symbol-function 'old-dump-array) (symbol-function 'compiler:fasd-array)))

(defun new-dump-array (array)
  (cond ((dbobject-p array )
	 (unless (relationp array)
	   (if *no-query-dump-dbobjects* (push array *no-query-dump-dbobjects*)
	       (cerror "Proceed to dump ~a."
		       "DUMP-ARRAY called on a dbobject (~a).~%~
                 This will work on your TI machine, but your code is not portable."
		       array)))
	 
	 (compiler:FASD-FORM
	   (cond ((relationp array)
		   `(relationp ',(relationnamep array)))
		  (t `(read-from-string
			,(let ((*print-escape* t))
			   (PrintDBObject array)))))
	   #+ignore `(read-from-string
	      ,(cond ((and bootstrapping (relationp array))
		      (format nil "#,(relationp '~A)"
			      (relationnamep array)))
		     (t (PrintDBObject array))))))
	#+ignore ;; ti does not use memodata objects
	((memodata-p array)
	 (when (done array)
	   (if *no-query-dump-memos* (push array *no-query-dump-memos*)
	     (cerror "Proceed to dump the memo cell."
		     "DUMP-ARRAY called on a loaded memo cell.~%~
                     This will work on your TI machine, but your code is not portable.")))
	 (compiler:FASD-FORM '(make-memodata)))
	(t (old-dump-array array))))

(setf (symbol-function 'compiler:fasd-array) (symbol-function 'new-dump-array))
)

;; **** this is just fixing a ti bug ...
#+ignore ; was ti
(global:advise si:copy-object-tree :around :skip-dbobjects nil
  (cond ((dbobject-p  (car global:arglist))  

	 (cerror "Just continue"
		 "Notify Don Cohen or Neil Goldman of this error before continuing,~
                  if possible~%~
                  Control was not expected to ever reach this point.~%~
                  You may safely resume from this error")
	 (car global:arglist))
	(t :do-it)))


; ---------------- the following was for an early version of allegro cl
#|
(setf comp::.random-forms-max. 0)
; that prevents compiling forms together - needed because the second might read
; a relation object defined by the first, in which case we need the first eval'd
; before the second is READ.

(unless (fboundp 'old-dump-array)
  (setf (symbol-function 'old-dump-array)
	(symbol-function #-sun4 'm68k::s-wfasl-lisp-vector
			 #+sun4 'hyperion::r-wfasl-lisp-vector)))
;; version of memo that passes the function as a parameter
(defun new-dump-array (array stream)
  (cond ((DBObject-p array)
	 (unless (relationp array)
	      (error  "DUMP-ARRAY called on a non-relation dbobject (~a)."
			 array))
	 (#+sun4 hyperion::r-wfasl-lispobj
	  #-sun4 m68k::s-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  . 
		  ;; note the period just before this line.  it is 
		  ;; important!
		  (read-from-string
		    ,(if bootstrapping
			 (format nil "#,(relationp '~A)"
				 (relationnamep array))
		       (let ((*print-escape* t))
				(PrintDBObject array)))))
		stream))
	((memodata-p array)
	 (when (done array)
	   (if *no-query-dump-memos* (push array *no-query-dump-memos*)
 	    (cerror "Proceed to dump the memo cell."
		   "DUMP-ARRAY called on a loaded memo cell.~%~
                 This will work in allegro, but your code is not portable.")))
	 ; it looks like we do indeed expect to compile memos to files
	 ; however they evidently are never dumped with done=T
	 (#-sun4 m68k::s-wfasl-lispobj 
          #+sun4 hyperion::r-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  . 
		  ;; note the period just before this line.  it is 
		  ;; important!
		  (make-memodata))
		stream))
	((loadmemodata-p array)
	 (#-sun4 m68k::s-wfasl-lispobj 
          #+sun4 hyperion::r-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  ,@(lm-form array))
		stream))
	(t (old-dump-array array stream))))


;; version of memo that stores function in memo cell
(defun new-dump-array (array stream)
  (cond ((DBObject-p array)
	 (unless (relationp array)
	      (error  "DUMP-ARRAY called on a non-relation dbobject (~a)."
			 array))
	 (#+sun4 hyperion::r-wfasl-lispobj
	  #-sun4 m68k::s-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  . 
		  ;; note the period just before this line.  it is 
		  ;; important!
		  (read-from-string
		    ,(if bootstrapping
			 (format nil "#,(relationp '~A)"
				 (relationnamep array))
		       (let ((*print-escape* t))
				(PrintDBObject array)))))
		stream))
	((memodata-p array)
	 (unless (done array)
	   (if *no-query-dump-memos* (push array *no-query-dump-memos*)
 	    (error "DUMP-ARRAY called on a loaded memo cell.")))
	 (#-sun4 m68k::s-wfasl-lispobj 
          #+sun4 hyperion::r-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  . 
		  ;; note the period just before this line.  it is 
		  ;; important!
		  (make-memodata :done (function ,(done array))))
		stream))
	((loadmemodata-p array)
	 (#-sun4 m68k::s-wfasl-lispobj 
          #+sun4 hyperion::r-wfasl-lispobj
		`(,comp::*eval-when-load-marker*
		  ,@(lm-form array))
		stream))
	(t (old-dump-array array stream))))

(setf (symbol-function #-sun4 'm68k::s-wfasl-lisp-vector
		       #+sun4 'hyperion::r-wfasl-lisp-vector)
      (symbol-function 'new-dump-array))
|#

